/*  ----------------------------------------------------------------------

    Copyright (C) 2000  Gabriel Gomiz  (gomita@bblanca.com.ar)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    ---------------------------------------------------------------------- */

    /* $Id$ */

#include "nomadIIpriv.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/ioctl.h>

#include <linux/usb.h>
#include <linux/usbdevice_fs.h>

#define MAX_SYNC_RETRIES   3

static int nomadII_get_file(NomadIIHandle, int, int, nomadII_file *);

static int retry_function(int (*func)(NomadIIHandle), NomadIIHandle nomadII, char* name)
{
	int retries = MAX_SYNC_RETRIES;
	
	while (retries--) {
		if (func(nomadII) == 0) {
			return 0;
		}
		usleep(100000);
	}

	nomadII_set_error("failed to get %s", name);
	return -1;
}

/* This function can be called as root, so must be watertight
   against buffer overruns, etc */
int nomadII_probe(NomadIIHandle nomadII)
{
#ifdef WITH_USBDEVFS
    int bus = 0, dev = 0, vendor = 0, prodid = 0;
    char buf[256];
    char tmp[PATH_MAX + 1];
    FILE *a;
    
    nomadII_reset_error();

    snprintf(tmp, PATH_MAX, "%s/devices", "/proc/bus/usb");
    tmp[PATH_MAX] = '\0';

    if ((a = fopen(tmp, "r")) == NULL) {
	nomadII_set_error("cannot open device: %s", strerror(errno));
        return -1;
    }

    while (fgets(buf, 255, a) != NULL)
    {
        if (buf[0] == 'T')
        {
            bus = strtol(&buf[8], NULL, 16);
            dev = strtol(&buf[46], NULL, 10);

        }
        else if (buf[0] == 'P')
        {
            vendor = strtol(&buf[11], NULL, 16);
            prodid = strtol(&buf[23], NULL, 16);

	    /* II and IIc: 0x1002, II MG: 0x4004 */
	    if ((vendor == 0x041E) && (prodid == 0x1002 || prodid == 0x4004))
            {
                snprintf(nomadII->dst_dev, PATH_MAX, "%s/%03d/%03d", "/proc/bus/usb", bus, dev);
                nomadII->dst_dev[PATH_MAX] = '\0';
		nomadII->bus = bus;
		nomadII->dev = dev;
		nomadII->vendor_id = vendor;
		nomadII->prod_id = prodid;
                return 0;
            }

        }

    }

    fclose(a);
    return -1;
#else
    int fd;

    fd = open("/dev/usb/nomadII",O_RDWR);
    if (fd < 0)
	    return -1;

    close(fd);
    return 0;
#endif
}

int nomadII_open(NomadIIHandle nomadII)
{
#ifdef WITH_USBDEVFS
	int interface = 0;
#endif

	nomadII_reset_error();

#ifdef WITH_USBDEVFS
	nomadII->fd = open(nomadII->dst_dev,O_RDWR);
#else
	nomadII->fd = open("/dev/usb/nomadII",O_RDWR);
#endif

	nomadII->nomadII_isopen = (nomadII->fd >= 0) ? 1 : 0;
	if (nomadII->fd < 0)
		return -1;

#ifdef WITH_USBDEVFS
	if (ioctl(nomadII->fd, USBDEVFS_CLAIMINTERFACE, &interface) == -1)
		return -1;
#endif

	return (nomadII->fd > 0) ? 0 : -1;
}

int nomadII_close(NomadIIHandle nomadII)
{
#ifdef WITH_USBDEVFS
	int interface = 0;
#endif
	CHECKOPEN(nomadII);

#ifdef WITH_USBDEVFS
	ioctl(nomadII->fd, USBDEVFS_RELEASEINTERFACE, &interface);
#endif		

	if (close(nomadII->fd) == 0) {
		nomadII->nomadII_isopen = 0;
		return 0;
	} else
		return -1;
} 

int nomadII_sync(NomadIIHandle nomadII)
{
	CHECKOPEN(nomadII);

	if (retry_function(nomadII_model, nomadII, "model information") != 0)
		return -1;
	if (retry_function(nomadII_memory, nomadII, "memory information") != 0)
		return -1;
	if (retry_function(nomadII_get_owner, nomadII, "owner information") != 0)
		return -1;
	if (retry_function(nomadII_get_clock, nomadII, "clock time") != 0)
		return -1;
	if (retry_function(nomadII_count_files, nomadII, "file count") != 0)
		return -1;
	if (retry_function(nomadII_get_files, nomadII, "file information") != 0)
		return -1;
	if (nomadII->has_radio)
		if (retry_function(nomadII_read_presets, nomadII, "radio presets") != 0)
			return -1;

	return 0;
}

int nomadII_prepare(NomadIIHandle nomadII)
{
	/* CONTROL 0x1A --> DOWNSTREAM (IN)             */
        /* CONTROL <-- UPSTREAM (IN) (length 0x01) (00) */
	/* RESET PIPES				        */
	/* CONTROL 0x10 --> DOWNSTREAM (IN)             */
        /* CONTROL <-- UPSTREAM (IN) (length 0x01) (00) */
	CHECKOPEN(nomadII);

	if (nomadII->prod_id != 0x4004) {
		if (nomadII_send_command(nomadII->fd,COMMAND_PREPARE,0x00,0x00) != 0)
			return -1;
	}

	if (nomadII_reset_pipes(nomadII->fd))
		return -1;

	if (nomadII_send_command(nomadII->fd,COMMAND_PREPARE2,0x00,0x00) != 0)
		return -1;
	return 0;
}

int nomadII_model(NomadIIHandle nomadII)
{
	/* CONTROL 0x0E --> DOWNSTREAM (IN)       */
	/* BULK --> DOWNSTREAM (IN) (length 0x74) */
	/* BULK <-- UPSTREAM   (IN) (data model)  */
	/* CONTROL <-- UPSTREAM (no data)         */

	BYTE buffer[0x74];         /* For bulk tranfer */
	int  ret, major, minor, subver;

	CHECKOPEN(nomadII);

	nomadII->has_radio = 1;

	/* Requesting model info */
	if (nomadII_send_command_nowait(nomadII->fd,COMMAND_MODEL,0x00,0x00) != 0)
		return -1;

	/* Reading model info */
	ret = nomadII_read(nomadII->fd,buffer,0x74);

	/* Always find result of COMMAND_MODEL */
	if (nomadII_command_wait(nomadII->fd) != 0 || ret != 0)
		return -1;

	/* Copy model string */
	strncpy(nomadII->model_info,(char *) &buffer[OFFSET_MODEL], STRING_SIZE);
	nomadII->model_info[STRING_SIZE-1] = '\0';

	if (nomadII->prod_id == 0x4004) {
		major = buffer[OFFSET_FW_MAJOR];
		minor = buffer[OFFSET_FW_MINOR_MG];
	} else {
		major = (buffer[OFFSET_FW_MAJOR] >> 4) + 1;
		minor = buffer[OFFSET_FW_MAJOR] & 0x0f;
	}
	subver = buffer[OFFSET_FW_SUBVER];
	
	snprintf(nomadII->fw_info, STRING_SIZE, "%d.%02d.%03d", major, minor, subver); 

	if (strstr(nomadII->model_info, "IIc") != NULL)
		nomadII->has_radio = 0;

	return 0;
}

int nomadII_get_clock(NomadIIHandle nomadII)
{
        /* CONTROL 0x11 --> DOWNSTREAM (IN) (no data)      */
        /* CONTROL <-- UPSTREAM (IN) (lenght 0x11) (clock) */

        BYTE buffer[0x11];

	CHECKOPEN(nomadII);

        /* Requesting owner info */
        if (nomadII_read_command(nomadII->fd,COMMAND_GET_CLOCK,0x00,0x00,0x11,buffer) != 0)
                return -1;

	/* Fill date structure with date buffer data */

	if (buffer[0x01] == 0x6c) {
		/* year offset of 7 allows two byte year 2000 (0x07d0) to be
		   compatible with new single byte format when reading year
	           0x07..0x56 -> 2000..2079, 1980..1999 -> 0x57..0x6a;
	           for new format buffer[0x01] has 0xd0 for y2000 else 0x6c */
		nomadII->dclock.tm_year = buffer[0x02] - 7;
		if (nomadII->dclock.tm_year < 80)
			nomadII->dclock.tm_year += 100;
	} else {
		memcpy(&nomadII->dclock.tm_year ,&buffer[0x01],sizeof(WORD));
		nomadII->dclock.tm_year -= 1900;
	}
	memcpy(&nomadII->dclock.tm_mon ,&buffer[0x03],sizeof(WORD));
	nomadII->dclock.tm_mon--; /* adjust to index from 0 */
	memcpy(&nomadII->dclock.tm_mday,&buffer[0x05],sizeof(WORD));
	memcpy(&nomadII->dclock.tm_wday,&buffer[0x07],sizeof(WORD));
	if (nomadII->dclock.tm_wday > 6)
		nomadII->dclock.tm_wday %= 7;
	memcpy(&nomadII->dclock.tm_hour,&buffer[0x09],sizeof(WORD));
	memcpy(&nomadII->dclock.tm_min ,&buffer[0x0B],sizeof(WORD));
	memcpy(&nomadII->dclock.tm_sec ,&buffer[0x0D],sizeof(WORD));

        return 0;
}

int nomadII_set_clock(NomadIIHandle nomadII)
{
        /* CONTROL 0x12 --> DOWNSTREAM (OUT) (lenght 0x10) (clock) */
        /* CONTROL <-- UPSTREAM (OUT)                              */
        /* CONTROL 0xF0 --> DOWNSTREAM (IN)  (ack)                 */
	/* CONTROL <-- UPSTREAM (IN) (lenght 0x01) (00)            */

        BYTE buffer[0x10];
	WORD year_aux,mon_aux;
	time_t local_clock;
	
	CHECKOPEN(nomadII);

	/* Get current clock on struct tm dclock */ 

	time(&local_clock);
	localtime_r(&local_clock,&nomadII->dclock);

	year_aux=1900+nomadII->dclock.tm_year;
	mon_aux=nomadII->dclock.tm_mon+1;
	if (nomadII->dclock.tm_wday == 0)
		nomadII->dclock.tm_wday += 7;
	memcpy(&buffer[0x00],&year_aux      ,sizeof(WORD));
	memcpy(&buffer[0x02],&mon_aux       ,sizeof(WORD));
	memcpy(&buffer[0x04],&nomadII->dclock.tm_mday,sizeof(WORD));
	memcpy(&buffer[0x06],&nomadII->dclock.tm_wday,sizeof(WORD));
	memcpy(&buffer[0x08],&nomadII->dclock.tm_hour,sizeof(WORD));
	memcpy(&buffer[0x0A],&nomadII->dclock.tm_min ,sizeof(WORD));
	memcpy(&buffer[0x0C],&nomadII->dclock.tm_sec ,sizeof(WORD));

        /* Setting device clock */
        if (nomadII_write_command(nomadII->fd,COMMAND_SET_CLOCK,0x00,0x00,0x10,buffer) != 0)
	   return -1;

	/*
	 * Command result
	 */
	if (nomadII_send_command(nomadII->fd,COMMAND_GET_RESULT,0x00,0x00) != 0)
	   return -2;	

        return 0;
}

int nomadII_get_owner(NomadIIHandle nomadII)
{
        /* CONTROL 0x1C --> DOWNSTREAM (IN) (no data)   */
        /* CONTROL <-- UPSTREAM (IN) (lenght 0x01) (00) */
        /* BULK --> DOWNSTREAM (IN) (lenght 0x40)       */
        /* BULK <-- UPSTREAM   (IN) (data owner)        */

        BYTE buffer[0x40];          /* For bulk transfer */

	CHECKOPEN(nomadII);

        /* Requesting owner info */
        if (nomadII_send_command(nomadII->fd,COMMAND_GET_OWNER,0x00,0x00) != 0)
                return -1;

        if (nomadII_read(nomadII->fd,buffer,0x40) != 0)
                return -1;

        /* Copy owner string */
        strncpy(nomadII->owner_info,(char *) &buffer[OFFSET_OWNER], STRING_SIZE);
	nomadII->owner_info[STRING_SIZE-1] = '\0';

        return 0;
}

int nomadII_set_owner(NomadIIHandle nomadII)
{
        /* CONTROL 0x1D --> DOWNSTREAM (OUT) (owner data) */
        /* CONTROL <-- UPSTREAM (OUT) (no data)           */
        /* CONTROL 0xF0 --> DOWNSTREAM (IN)  (ack)        */
	/* CONTROL <-- UPSTREAM (IN) (lenght 0x01) (00)   */

        BYTE buffer[0x40];

	CHECKOPEN(nomadII);

        /* Copy owner string */
        strncpy((char *) &buffer[OFFSET_OWNER],nomadII->owner_info,0x40);

        /* Setting owner info */
        if (nomadII_write_command(nomadII->fd,COMMAND_SET_OWNER,0x00,0x00,0x40,buffer) != 0)
                return -1;

	if (nomadII_send_command(nomadII->fd,COMMAND_GET_RESULT,0x00,0x00) != 0)
		return -1;	

        return 0;
}

int nomadII_memory(NomadIIHandle nomadII)
{
        /* CONTROL 0x0D --> DOWNSTREAM (IN) (no data)            */
        /* CONTROL <-- UPSTREAM (IN) (lenght 0x11) (data memory) */

        BYTE buffer[0x11];

	CHECKOPEN(nomadII);

        if (nomadII_read_command(nomadII->fd,COMMAND_MEMORY,0x00,0x00,0x11,buffer) != 0)
                return -1;

        memcpy((void *)&nomadII->mem_total[INTERNAL],
        	&buffer[OFFSET_MEM_INTERNAL_TOTAL],sizeof(DWORD));
        memcpy((void *)&nomadII->mem_free[INTERNAL],
        	&buffer[OFFSET_MEM_INTERNAL_FREE],sizeof(DWORD));
        memcpy((void *)&nomadII->mem_total[EXTERNAL],
        	&buffer[OFFSET_MEM_EXTERNAL_TOTAL],sizeof(DWORD));
        memcpy((void *)&nomadII->mem_free[EXTERNAL],
        	&buffer[OFFSET_MEM_EXTERNAL_FREE],sizeof(DWORD));

	nomadII->present[INTERNAL] = nomadII->mem_total[INTERNAL] > 0;
	nomadII->present[EXTERNAL] = nomadII->mem_total[EXTERNAL] > 0;
	nomadII->both_present = nomadII->present[INTERNAL] && nomadII->present[EXTERNAL];

        return 0;
}

int nomadII_count_files(NomadIIHandle nomadII)
{
        /* CONTROL 0x01 val 0x01 --> DOWNSTREAM (OUT) (no data)     */
        /* CONTROL <-- UPSTREAM (OUT)                               */
        /* CONTROL 0x01 --> DOWNSTREAM (IN)                         */
        /* CONTROL <-- UPSTREAM (IN)  (length 0x03) (data 00 nn 00) */

        BYTE buffer[3];         /* temporary buffer - upstream data */

	CHECKOPEN(nomadII);

	if (nomadII->present[INTERNAL]) {
        	if (nomadII_write_command(nomadII->fd,COMMAND_FILE_COUNT,INTERNAL,0x00,0x00,NULL) != 0)
                	return -1;

        	if (nomadII_read_command(nomadII->fd,COMMAND_FILE_COUNT,0x00,0x00,0x03,(BYTE*)&buffer) != 0)
                	return -1;

        	memcpy(&nomadII->bank_count[INTERNAL],&buffer[OFFSET_FILE_COUNT],sizeof(WORD));
	} else {
		nomadII->bank_count[INTERNAL] = 0;
	}

	if (nomadII->present[EXTERNAL]) {
        	if (nomadII_write_command(nomadII->fd,COMMAND_FILE_COUNT,EXTERNAL,0x00,0x00,NULL) != 0)
                	return -1;

        	if (nomadII_read_command(nomadII->fd,COMMAND_FILE_COUNT,0x00,0x00,0x03,(BYTE*)&buffer) != 0)
                	return -1;

        	memcpy(&nomadII->bank_count[EXTERNAL],&buffer[OFFSET_FILE_COUNT],sizeof(WORD));
	} else {
		nomadII->bank_count[EXTERNAL] = 0;
	}

	nomadII->file_count = nomadII->bank_count[INTERNAL] + nomadII->bank_count[EXTERNAL];

	if (nomadII->file_count > MAX_FILES) {
		int truncate = nomadII->file_count - MAX_FILES;
	
		if (nomadII->bank_count[EXTERNAL] > 0) {
			if (truncate > nomadII->bank_count[EXTERNAL]) {			    
				truncate -= nomadII->bank_count[EXTERNAL];
				nomadII->bank_count[EXTERNAL] = 0;
			} else {
				nomadII->bank_count[EXTERNAL] -= truncate;
			}
		}
			
		if (truncate > 0 && nomadII->bank_count[INTERNAL] > 0) {
			if (truncate > nomadII->bank_count[INTERNAL]) {			    
				truncate -= nomadII->bank_count[INTERNAL];
				nomadII->bank_count[INTERNAL] = 0;
			} else {
				nomadII->bank_count[INTERNAL] -= truncate;
			}
		}

		assert(truncate == 0);

		nomadII->file_count = MAX_FILES;
		nomadII_set_error("error: total number of files (%d) exceeds hardcoded maximum (%d).",
			nomadII->file_count, MAX_FILES);
		return -1;
	}

        return 0;
}

static int nomadII_get_file(NomadIIHandle nomadII, int bank, int i, nomadII_file *f)
{
	/* CONTROL 0x02 val 0x01 idx #i --> DOWNSTREAM (IN)                      */
	/* CONTROL                      <-- UPSTREAM (IN) (lenght 0x02) (xx yy)  */
	/* BULK                         --> DOWNSTREAM (IN) (lenght 0x215)       */
	/* BULK                         <-- UPSTREAM   (IN) (data chars & name)  */

	BYTE status[2] = { 0x00, 0x00 };
	BYTE buffer[0x215];
	WORD  mdate, mtime;

	CHECKOPEN(nomadII);

	/* Requesting file info */
	if (nomadII_write_command(nomadII->fd,COMMAND_FILE_INFO,bank,i,0x02,(BYTE*)&status) != 0)
		return -1;
	
	if (nomadII_read(nomadII->fd,buffer,0x215) != 0)
		return -1;

	nomadII_set_file_name((char *)&buffer[OFFSET_FILE_NAME],f->file_name);
	nomadII_set_file_ext((char *)&buffer[OFFSET_FILE_EXT],f->file_ext);
	nomadII_set_long_file_name((char *)&buffer[OFFSET_LONG_FILE_NAME],f->long_file_name);
	f->bank = bank;

	/* Copy file size */		
	memcpy(&(f->file_size),&buffer[OFFSET_FILE_SIZE],sizeof(DWORD));

	/* Copy date/time */		
	memcpy(&mdate,&buffer[OFFSET_FILE_DATE],sizeof(WORD));
	memcpy(&mtime,&buffer[OFFSET_FILE_TIME],sizeof(WORD));

	switch(buffer[OFFSET_FILE_TYPE]) {
		case 0x07 : f->type = FILE_AUX;     break;
		case 0x20 : f->type = FILE_ANY;     break;
		default   : f->type = FILE_UNKNOWN; break;
	}

	f->mtime = nomadII_date_dos2unix(mdate, mtime);

	return 0;
}

int nomadII_get_files(NomadIIHandle nomadII)
{
	int i, j, bank;

	CHECKOPEN(nomadII);

	if (nomadII->file_count > MAX_FILES) {
		nomadII->file_count = MAX_FILES;
		nomadII_set_error("error: total number of files (%d) exceeds hardcoded maximum (%d).",
			nomadII->file_count, MAX_FILES);
		return -1;
	}

	for (bank = 0, j = 0; bank < MEMORY_BANKS; bank++) {
		int aux_files = 0;

		for (i = 0; i < nomadII->bank_count[bank]; i++) {
			if (nomadII_get_file(nomadII, bank,i,&nomadII->files[j]) != 0) {
				return -1;
			} else {
				/* ignore .AUX files */
				if (nomadII->files[j].type == FILE_AUX)
					aux_files++;
				else
					j++;
			}
		}
		nomadII->bank_count[bank] -= aux_files;
		nomadII->file_count       -= aux_files;
	}
	return 0;
}

int nomadII_format_flash(NomadIIHandle nomadII, int bank) 
{
	/* CONTROL 0x0F val 0x201 --> DOWNSTREAM (IN)   */
	/* CONTROL <-- UPSTREAM (IN) (lenght 0x01) (00) */

	CHECKOPEN(nomadII);

	return nomadII_send_command(nomadII->fd,COMMAND_FORMAT, 0x0200 | bank, 0x00);	
}


int nomadII_read_presets(NomadIIHandle nomadII)
{
        /* CONTROL 0x1E --> DOWNSTREAM (IN)                       */
	/* CONTROL <-- UPSTREAM (IN) (lenght 0x41) (presets data) */

	int i;
	unsigned short fm1,fm2;

	BYTE buffer[0x41];

	CHECKOPEN(nomadII);

	if (nomadII_read_command(nomadII->fd,COMMAND_GET_RADIO_PRESETS,0x00,0x00,0x41,buffer) != 0)
		return -1;

	if (buffer[0] == 0) {
		for (i=0; i<=31; i++) {
			memcpy(&fm1,&buffer[2*i+1],2);
			fm2 = fm1 % 100;
			fm1 = fm1 / 100;

			nomadII->radios[i].freq    = fm1;
			nomadII->radios[i].subfreq = fm2;
		}
    
		return 0;

	}
	
	return -1;
}

int nomadII_write_presets(NomadIIHandle nomadII)
{
        /* CONTROL 0x1F --> DOWNSTREAM (IN) (lenght 0x40) (presets data) */
	/* CONTROL <-- UPSTREAM (OUT)                                    */
	/* CONTROL 0xF0 --> DOWNSTREAM (IN)                              */
	/* CONTROL <-- UPSTREAM (IN) (lenght 0x01) (00)                  */

	int i;
	short fm;

	BYTE buffer[64];

	CHECKOPEN(nomadII);

	for (i=0; i<=31; i++) {

		fm = (nomadII->radios+i)->freq * 100 + (nomadII->radios+i)->subfreq;

		memcpy(&buffer[2*i],&fm,2);
      
	}

	if (nomadII_write_command (nomadII->fd,COMMAND_SET_RADIO_PRESETS,0x00,0x00,0x40,buffer)<0)
		return -1;

	if (nomadII_send_command(nomadII->fd,COMMAND_GET_RESULT,0x00,0x00) != 0) {
		return -1;	
	}

	return 0;
}

int nomadII_delete(NomadIIHandle nomadII, int number) 
{
	int i;
	char *file_name = nomadII->files[number].file_name;
	char *file_ext  = nomadII->files[number].file_ext;
	int   bank      = nomadII->files[number].bank;

	BYTE command[] = { 0x00, 0x00, 0x00,
			   0x00, 0x00, 0x00,
			   0x00, 0x00, 0x00,
			   0x00, 0x00 };

	CHECKOPEN(nomadII);

	for (i = 0; i < 8; i++)
		command[i] = *(file_name+i);
	for (i = 0; i < 3; i++)
		command[i+8] = *(file_ext+i);

	/* CONTROL 0x08 --> DOWNSTREAM (IN)  (no data) */
	/* CONTROL <-- UPSTREAM (IN)                   */

	if (nomadII_write_command(nomadII->fd,COMMAND_DELETE_FILE,bank,0x00,0x00,NULL) != 0)
		return -1;

	if (nomadII_write(nomadII->fd,command,sizeof(command)) != 0)
		return -1;

	/* CONTROL 0xF0 --> DOWNSTREAM (IN)             */
	/* CONTROL <-- UPSTREAM (IN) (lenght 0x01) (00) */
	if (nomadII_send_command(nomadII->fd,COMMAND_GET_RESULT,0x00,0x00) != 0)
		return -1;

	if (nomadII_memory(nomadII) != 0)
		return -1;

	return 0;	
}
