/*  ----------------------------------------------------------------------

    Copyright (C) 2000  Gabriel Gomiz  (gomita@bblanca.com.ar)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    ---------------------------------------------------------------------- */

    /* $Id$ */

#include "nomadIIpriv.h"

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include <asm/byteorder.h>
#include <linux/usb.h>
#include <linux/usbdevice_fs.h>

typedef struct {
	__u8 requesttype;
	__u8 request;
	__u16 value;
	__u16 index;
	__u16 length;
} devrequest __attribute__ ((packed));

#ifdef PACKET_DUMP
#define MAX_DUMP_SIZE 0x0300
#endif

#ifdef PACKET_DUMP
int nomadIIdebug = 0;
#endif

int nomadII_read_command(int fd, int req, int val, int idx, int len, BYTE *buffer)
{
	int len_read;
	CHECKFD();
	
	len_read = nomadII_ctl_msg(fd, USB_DIR_IN, req, val, idx, len, buffer);
#ifdef PACKET_DUMP
	if (nomadIIdebug > 0 && len_read >= 0 && len_read != len)
		fprintf(stderr, "WARNING: nomadII_read_command read %d bytes - expected to read %d bytes\n",
				len_read, len);
#endif
	return (len_read >= 0) ? 0 : -1;
}

int nomadII_write_command(int fd, int req, int val, int idx, int len, BYTE *buffer)
{
	int len_written;
	CHECKFD();
	
	len_written = nomadII_ctl_msg(fd, USB_DIR_OUT, req, val, idx, len, buffer);
#ifdef PACKET_DUMP
	if (nomadIIdebug > 0 && len_written >= 0 && len_written != len)
		fprintf(stderr, "WARNING: nomadII_write_command wrote %d bytes - expected to write %d bytes\n",
				len_written, len);
#endif
	return (len_written >= 0) ? 0 : -1;
}

int nomadII_read(int fd, BYTE *buffer, int count)
{
	int result;
  
	CHECKFD();

#ifdef WITH_USBDEVFS
	result = nomadII_bulk_msg(fd,USB_DIR_IN, count, buffer);
#else
	result = read(fd,buffer,count);
	if (result == count)
		result = 0;
	else {
#ifdef PACKET_DUMP
		if (nomadIIdebug > 0 && result >= 0 && result != count)
			fprintf(stderr, "WARNING: nomadII_read read %d bytes - expected to read %d bytes\n",
				result, count);
#endif
		errno  = EIO;
		result = -1;
	}
#endif

#ifdef PACKET_DUMP
	if (nomadIIdebug > 1 || (nomadIIdebug > 0 && result < 0)) {
		fprintf(stdout,"* nomadII_read(%p,0x%04x) = 0x%02x\n", buffer, count, result);
		nomadII_hex_dump(buffer, count);
	}
#endif
       
        return result;
}

int nomadII_write(int fd, BYTE *buffer, int count)
{
	int result;

	CHECKFD();

#ifdef WITH_USBDEVFS
	result = nomadII_bulk_msg(fd, USB_DIR_OUT, count, buffer);
#else
	result = write(fd,buffer,count);
	if (result == count) 
		result = 0;
	else {
#ifdef PACKET_DUMP
		if (nomadIIdebug > 0 && result >= 0 && result != count)
			fprintf(stderr, "WARNING: nomadII_write wrote %d bytes - expected to write %d bytes\n",
				result, count);
#endif
		errno  = EIO;
		result = -1;
	}
#endif

#ifdef PACKET_DUMP
	if (nomadIIdebug > 1 || (nomadIIdebug > 0 && result < 0)) {
		fprintf(stdout,"* nomadII_write(%p,0x%04x) = 0x%02x\n", buffer, count, result);
		nomadII_hex_dump(buffer, count);
	}
#endif

        return result;
}

int nomadII_send_command(int fd, int req, int val, int idx)
{
	BYTE status;
	int  result;

	CHECKFD();

	result = nomadII_ctl_msg(fd, USB_DIR_IN, req, val, idx, 0x01, &status);

#ifdef PACKET_DUMP
	if (nomadIIdebug > 0 && result >= 0 && result != 1)
		fprintf(stderr, "WARNING: nomadII_send_command result != 1 indicates nomadII_send_command"
				"         should not have been used: req=0x%02x result=0x%02x\n", req, result);
#endif
	if (result >= 0 && status == 0)
		errno = 0;
	else
		switch(status) {
			case RESULT_ILLEGAL_FILENAME:
				errno = ENOENT;
				break;
			case RESULT_ILLEGAL_ARGUMENT:
				errno = EINVAL;
				break;
			case RESULT_ILLEGAL_FREQUENCY:
			case RESULT_ILLEGAL_FREQ_DECIMAL:
				nomadII_set_error("illegal frequency");
				errno = EINVAL;
				break;
			default:
#ifdef PACKET_DUMP
				if (nomadIIdebug > 0) {
					fprintf(stdout,"* (0x%02x,0x%02x,0x%04x,0x%04x) = 0x%02x\n",
						USB_DIR_IN,req,val,idx,result);
					fprintf(stderr,"status = 0x%02x\n", status);	
				}	
#endif
				errno = EIO;
		}

        return status;
}

#ifdef WITH_USBDEVFS
static struct usbdevfs_urb uurb;
#endif
static int uurb_inuse = 0; /* designed to give partial protection only */

int nomadII_send_command_nowait(int fd, int req, int val, int idx)
{
#ifndef WITH_USBDEVFS
	struct NomadII_cmd msg;
#endif

	devrequest  dr;
	int	    result;
	__u16	    uval = val;
	__u16	    uidx = idx;

	CHECKFD();

	if (uurb_inuse) {
		errno = EBUSY;
		return -1;
	}

	dr.requesttype = USB_TYPE_VENDOR | USB_DIR_IN | USB_RECIP_OTHER;
	dr.request     = req;
	dr.value       = __cpu_to_le16p(&uval);
	dr.index       = __cpu_to_le16p(&uidx);
	dr.length      = 0;

	uurb_inuse = 1;

#ifdef WITH_USBDEVFS
	uurb.type          = USBDEVFS_URB_TYPE_CONTROL;
	uurb.endpoint      = USB_DIR_IN | USB_ENDPOINT_XFER_CONTROL;	
	uurb.flags         = 0;
	uurb.buffer        = &dr;
	uurb.buffer_length = 8;	
	uurb.start_frame   = 0;
	uurb.signr         = 0;
	uurb.number_of_packets = 0;
	
	result = ioctl(fd, USBDEVFS_SUBMITURB, &uurb);
#else
	msg.requesttype = USB_TYPE_VENDOR | USB_DIR_IN | USB_RECIP_OTHER;
	msg.request     = req;
	msg.value       = val;
	msg.index       = idx;
	msg.length      = 8;
	msg.data        = &dr;

	result = ioctl(fd, NOMADII_ASYNC_COMMAND, &msg);
#endif

#ifdef PACKET_DUMP
	if (nomadIIdebug > 1 || (nomadIIdebug > 0 && result < 0)) {
		fprintf(stdout,"* (0x%02x,0x%02x,0x%04x,0x%04x) = 0x%02x\n",
			dr.requesttype,req,val,idx,result);
	}
#endif

	return result;
}

int nomadII_command_wait(int fd)
{
#ifdef WITH_USBDEVFS
	struct usbdevfs_urb* ruurb;
#endif

	CHECKFD();

	if (!uurb_inuse) {
		errno = EAGAIN;
		return -1;
	}

	uurb_inuse = 0;

#ifdef WITH_USBDEVFS
	if (ioctl(fd, USBDEVFS_REAPURB, &ruurb) != 0)
		return -1;

#ifdef PACKET_DUMP
	if (nomadIIdebug && ruurb != &uurb)
		fprintf(stderr, "warning: reaped URB did not have expected address of %x: %x\n",
			(int)&uurb, (int)ruurb);
#endif

	if (ruurb->status != 0) {
		errno = ruurb->status;
		return -1;
	}
#else
	return ioctl(fd, NOMADII_ASYNC_COMPLETE, NULL);
#endif

	return 0;
}

int nomadII_reset_pipes(int fd)
{
	unsigned int p;
#ifndef WITH_USBDEVFS
	struct NomadII_cmd cmd;
#endif
	p = NOMADII_BULK_IN | USB_DIR_IN;
	
	CHECKFD();

#ifdef WITH_USBDEVFS
	if  (ioctl (fd, USBDEVFS_CLEAR_HALT, &p) == -1)
		return -1;

	p = NOMADII_BULK_IN | USB_DIR_OUT;
	return ioctl (fd, USBDEVFS_CLEAR_HALT, &p);
#else
	cmd.timeout = 50; cmd.requesttype = 0; cmd.request = 0;
	cmd.value   = 0;  cmd.index       = 0; cmd.length  = 0;
	cmd.data    = NULL;

	return ioctl (fd, NOMADII_RESET_PIPES, &cmd);	
#endif
}

int nomadII_ctl_msg(int fd, int reqtype, int req, int val, int idx, int len, BYTE *d)
{
	int result;
#ifdef WITH_USBDEVFS
	struct usbdevfs_ctrltransfer msg;
#else
	struct NomadII_cmd msg;
#endif
	
	CHECKFD();

	/* Rest of parameters */
	msg.requesttype = USB_TYPE_VENDOR | reqtype | USB_RECIP_OTHER;
	msg.request     = req;
	msg.value       = val;
	msg.index       = idx;
	msg.length      = len;
	msg.data        = d;

	/* The timeout depends on the command */
        switch (req) {

        case COMMAND_MEMORY:
        case COMMAND_GET_RADIO_PRESETS:
        case COMMAND_SET_RADIO_PRESETS:
                msg.timeout = NOMADII_TIMEOUT_MEDIUM;
                break;

        case COMMAND_SEND_BLOCK:
	case COMMAND_RCV_BLOCK:
	case COMMAND_SET_CLOCK:
	case COMMAND_END_TRANSFER:
	case COMMAND_FORMAT:
	case COMMAND_GET_RESULT:
                msg.timeout = NOMADII_TIMEOUT_LARGE;
                break;

        default:
                msg.timeout = NOMADII_TIMEOUT_SMALL;
                break;
        }


#ifdef WITH_USBDEVFS
	result = ioctl (fd, USBDEVFS_CONTROL, &msg);
#else
	switch (reqtype) {

	case USB_DIR_IN:
		result = ioctl (fd, NOMADII_RECV_COMMAND, &msg);
		break;

	case USB_DIR_OUT:
		result = ioctl (fd, NOMADII_SEND_COMMAND, &msg);
		break;

	default:
               result = -1;
	}	
#endif

#ifdef PACKET_DUMP
	if (nomadIIdebug > 1 || (nomadIIdebug > 0 && result < 0)) {
		fprintf(stdout,"* (0x%02x,0x%02x,0x%04x,0x%04x,0x%04x,%p) = 0x%02x\n",
			reqtype,req,val,idx,len,d,result);
		nomadII_hex_dump(d, len);
	}
#endif

	return result;
}

#ifdef WITH_USBDEVFS
int nomadII_bulk_msg(int fd, int direction, int len, BYTE *data) {

	struct usbdevfs_bulktransfer bulk;
	int retries,result=0;

	CHECKFD();

	bulk.ep = NOMADII_BULK_IN | direction;
	bulk.timeout = BULK_TIMEOUT;
	bulk.len = len;
	bulk.data = data;

	retries = BULK_MAX_RETRIES;

	while (retries--) {
		if ((result = ioctl(fd, USBDEVFS_BULK, &bulk)) == -1
		    && errno == ETIMEDOUT) {
			/* NAK so hold for a while */		    
#ifdef PACKET_DUMP
			if (nomadIIdebug)
				printf("nomadII_bulk_msg timed out, trying again\n");
#endif
			usleep(BULK_INTERVAL);
			continue;
		}
		else
			break;
	}
    
	return (result == len) ? 0 : -1;
}
#endif
